import AvMenu from './AvMenu.js';
import SoundMeter from './SoundMeter.js';

let pc = undefined;
let audioCanvas = undefined;
let audioVisualizer = undefined;

let soundMeter = undefined;
let soundMeterInterval = undefined;

window.AudioContext = window.AudioContext || window.webkitAudioContext || false;


//state表示当前的状态，其中publishState是发布状态，可选的值包括 ready,pending,publishing
let state = {

	publishState: "ready",
	constraints : {
					"video":{"width":{"min":640,"ideal":1280,"max":1920},"height":{"min":360,"ideal":720,"max":1080},"frameRate":{"max":30}},
					"audio":true
				},
	localStream:undefined,
	videoElementPublish: undefined,
	selectedCam:undefined,
	selectedMic:undefined,
	cameras:[],
	microphones:[],
	useSoundMeter: true,
	muted: false,
	video: true,
	audioEnabled: true,
	videoEnabled: true
	
}	


$(document).ready(() => {
	
	console.log("ready");
	state.videoElementPublish = document.getElementById('publisher-video');
	audioCanvas = document.getElementById('audioCanvas');

	let protocol = window.location.protocol;
	let host = window.location.hostname;
	let port = window.location.port;
	
	let base_url = protocol + "//"+ host +":"+ port +"/" ;
	if (port == "")
	{
		base_url = protocol + "//"+ host +"/" 
	}
	
	$("#url").val(base_url);
	
	
	initListener();	
	loadLocalStream();		
	
});


const initListener = async() =>
{

	$("#publish-toggle").click((e) => {
		
		if (state.publishState == "publishing"){
			stop();
			clearLogs();
		}else if (state.publishState == "pending"){
			return;
		}
		else 
		{
			try {
			    start();
			}
			catch(e){
			  log(e);
			}
		}
		
	}
	);
		
	$("#frameSize").change(() => {
     		
		let newConstraints = JSON.parse(JSON.stringify(state.constraints));
		 
		let resolution = $("#frameSize option:selected").val();
		if (resolution !== 'default')
		{
			let width_height_size = resolution.split("x");
			newConstraints.video["width"] = { exact: width_height_size[0] };
			newConstraints.video["height"] = { exact: width_height_size[1] };
		
		}else
		{
			newConstraints.video["width"] = { min: 640, ideal: 1280, max: 1920 };
			newConstraints.video["height"] = { min: 360, ideal: 720, max: 1080 };
		}
		//如果设置发生了变化
		if (newConstraints != null && JSON.stringify(state.constraints) !== JSON.stringify(newConstraints))
		{
			//覆盖现有的Contraints
			state.constraints = JSON.parse( JSON.stringify(newConstraints));
			
			//作用在MediaStream上
			loadLocalStream();
			
		}

	});	
	
	$("#mute-toggle").click((e) => {
      e.preventDefault()
      state.muted = !state.muted
      if(state.muted) {
        $("#mute-off").css("display", "none");
        $("#mute-on").css("display", "inline");
        
		//将localStream中的音频轨道停止
		state.audioEnabled = false;
		setEnabled("audio",false);
      }else{
        $("#mute-on").css("display", "none");
        $("#mute-off").css("display", "inline");
        state.audioEnabled = true;
		setEnabled("audio",true);
		
      }
    });
	
	
	$("#camera-toggle").click((e) => {
      e.preventDefault()
      state.video = !state.video
      if(state.video) {
        $("#video-on").css("display", "none");
        $("#video-off").css("display", "inline");
		state.videoEnabled = true;
		setEnabled("video",true);
        
      }else{
        $("#video-off").css("display", "none");
        $("#video-on").css("display", "inline");
		state.videoEnabled = false;
		setEnabled("video",false);
       
      }
    });
}

const loadLocalStream = async () =>
{
	if (state.localStream == null)
	{
		await getUserMedia();	
	}

	await getDevices();
	
	if (state.localStream != null)
	{
		applyConstraints(state.localStream,state.constraints).then(log("apply Constraints success")).catch(log);
	}
	
}
	

const start = () =>{	
	
	clearLogs();
	
	let signal_url = document.getElementById("url").value;
	if (signal_url == "")
	{
		$("#url").css("border-color","#FF0000");
		return;
	}
	$("#url").css("border-color","#CCCCCC");	
	
	state.publishState = "pending";
		
    pc = new RTCPeerConnection({
        iceServers: [],
        rtcpMuxPolicy: 'require',
        bundlePolicy: 'max-bundle',
    })
    pc.oniceconnectionstatechange = e => log(pc.iceConnectionState)
    pc.onicecandidate = event => {
        if (event.candidate === null) {
            log("gather done " + JSON.stringify(pc.localDescription));
			
            fetch(signal_url, { 
                method: "POST",
                body: JSON.stringify({
                    sdp: pc.localDescription
                }),
                headers: {
                    "Content-type": "application/json; charset=UTF-8"
                }
            })
            .then((response) => response.json())
            .then((json) => {
                log("--- got response")
                log(JSON.stringify(json))
				try {
                    pc.setRemoteDescription(new RTCSessionDescription(json.data.sdp))
					state.publishState = "publishing";
					$("#publish-toggle").text("停止推流");
                } catch (e) {
                    state.publishState = "ready";
					log(e)
                }
            })
			.catch( (e)=>{
					log(e);
					state.publishState = "ready";
			});
			
        }
    }
    
	let localTracks = state.localStream.getTracks();
	for (let localTrack in localTracks) {
	  let sender = pc.addTrack(localTracks[localTrack], state.localStream);
	}
 
	pc.createOffer()
		.then(d => pc.setLocalDescription(d))
        .catch(log);
		
	  
}

const stop = () =>{

	if (pc != null)
	{
		pc.close();
	}	

	pc = undefined;
	$("#publish-toggle").text("启动推流");
	state.publishState = "ready";
	log("connection closed");
}




//获得媒体并授权
const getUserMedia = (mediaKind) =>
{
  mediaKind = mediaKind || 'both';

  return new Promise((resolve,reject) =>
  {
	
	let savedAudioTracks = [];
    let savedVideoTracks = [];
	
	if (state.localStream != null)
    {
      savedAudioTracks = state.localStream.getAudioTracks();
      savedVideoTracks = state.localStream.getVideoTracks();
    }

    if (mediaKind !== 'video')
    {
      if (soundMeter != null)
      {
        soundMeter.stop();
      }
      if (soundMeterInterval != null)
      {
        clearInterval(soundMeterInterval);
      }
    }
	
	if (state.videoElementPublish == null)
    {
      reject({message:"videoElementPublish not set"});
    }
		
	
	async function getUserMediaSuccess(stream)
    {
      if (mediaKind === 'audio' && savedVideoTracks.length > 0)
      {
        let videoTracksToRemove = stream.getVideoTracks();
        for(let i in videoTracksToRemove)
        {
          stream.removeTrack(videoTracksToRemove[i]);
        }
        stream.addTrack(savedVideoTracks[0]);
      }
      else if (mediaKind === 'video' && savedAudioTracks.length > 0)
      {
        let audioTracksToRemove = stream.getAudioTracks();
        for(let j in audioTracksToRemove)
        {
          stream.removeTrack(audioTracksToRemove[j]);
        }
        stream.addTrack(savedAudioTracks[0]);
      }
	  state.localStream = stream;
     // let newState = {stream:stream};
      if (mediaKind !== 'audio' && state.isScreenSharing)
      {
        for(let k in savedVideoTracks)
        {
          savedVideoTracks[k].stop();
        }
        state['isScreenSharing'] = false;
      }
      try
      {
        state.videoElementPublish.srcObject = stream;
      }
      catch (error)
      {
        console.log('getUserMediaSuccess: error connecting stream to videoElementPublish, trying createObjectURL');
        console.log(error);
        state.videoElementPublish.src = window.URL.createObjectURL(stream);
        //state['ready'] = true;
      }
	  
	  state.publishState = "ready";
	  	
      try
      {
        if (mediaKind !== 'video' && window.AudioContext && state.useSoundMeter)
        {

		  //simon代码中的音量可视化
          audioVisualizer = new StreamVisualizer(stream, audioCanvas);
		  audioVisualizer.start();
		
		
		  let audioContext = new AudioContext();
          let soundMeter = new SoundMeter(audioContext);
          soundMeter.connectToSource(stream, function(e) {
            if (e) {
              console.log(e);
              return;
            }
            soundMeterInterval = setInterval(() => {
              let soundVal = state.audioEnabled ? soundMeter.instant.toFixed(2) : 0 ;
              if (onSoundMeter != null){
                onSoundMeter(soundVal);
              }
            }, 200);
          });
        }
      }
      catch (error2)
      {
        console.log('getUserMediaSuccess: error creating audio meter');
        console.log(error2);
      }
     
      resolve(state);
    }	
	

	if (navigator.mediaDevices.getUserMedia)
    {
        navigator.mediaDevices.getUserMedia(state.constraints)
        .then(getUserMediaSuccess)
        .catch(log);
    }
    else if (navigator.getUserMedia)
    {
       navigator.getUserMedia(state.constraints, getUserMediaSuccess, (error) => {
        log(error);
        reject(error);
      });
    }
    else
    {
      reject();
    }
	
  });
	
}



//获取设备列表
const getDevices = () =>
{
  return new Promise((resolve,reject) =>
  {
	navigator.mediaDevices.enumerateDevices().then((devices) =>
    {
     
      let cameras = [];
      let microphones = [];
      for(let i = 0; i < devices.length; i++){
        if(devices[i].kind === 'videoinput'){
          if (cameras.length === 0) {
            state.constraints.video = Object.assign({},state.constraints.video,{deviceId: devices[i].deviceId});
          }
          cameras.push(devices[i]);
		}else if (devices[i].kind === 'audioinput') {
          if (microphones.length === 0) {
            state.constraints.audio = Object.assign({},state.constraints.audio,{deviceId: devices[i].deviceId});
          }
          microphones.push(devices[i]);
        }
      }
      state.cameras = cameras;
	  state.microphones = microphones;
      AvMenu.init(cameras,microphones, onAvMenuCameraChanged, onAvMenuMicrophoneChanged);
	  resolve(state);
    }).catch(
      (e) => {
        console.log("unable to detect AV devices: " + e);
         reject(e);
      }
    ); 
	
  });
}  
	
const applyConstraints = (stream, constraints) => {
  let promises = [];
  let audioTracks = stream.getAudioTracks();
  let videoTracks = stream.getVideoTracks();
  
  for (let a in audioTracks)
  {
    promises.push(audioTracks[a].applyConstraints(constraints['audio']));
  }
  for (let v in videoTracks)
  {
    console.log(JSON.stringify(constraints['video']));
	promises.push(videoTracks[v].applyConstraints(constraints['video']));
  }
  
  
  return Promise.all(promises);
}

 // sound meter
const onSoundMeter = (level) => {
    // map level onto the rising quadrant shape of a circle to exaggerate the sound meter gradient
    let shiftLevel = level - 1;
    let levelCirc = Math.round(100 * Math.sqrt( 1 - (shiftLevel * shiftLevel) ));
    $("#mute-toggle").css("background-image","linear-gradient(white "+(100-levelCirc)+"%, lime)");
};

const onAvMenuCameraChanged = (cameraId) => {
    if (state.selectedCam !== cameraId)
    {
      state.selectedCam = cameraId;
      setCamera(cameraId);
    }
}

const onAvMenuMicrophoneChanged = (microphoneId) => {
	if (state.selectedMic !== microphoneId)
	{
	  state.selectedMic = microphoneId;
	  setMicrophone(microphoneId);
	}
}

const setEnabled = (trackKind, enabled) => {
 
  if (state.localStream != null && state.localStream.getTracks != null)
  {
    state.localStream.getTracks().map((track) => {
      if (track.kind === trackKind) {
        track.enabled = enabled;
      }
    });
  }
}


const setMicrophone = (id) =>
{
	console.log("setMicrophone: " + id);

	if (typeof state.constraints.audio === 'boolean')
	{
		state.constraints.audio = {};
	}
	state.constraints.audio = Object.assign({},state.constraints.audio,{deviceId: id});
	getUserMedia('audio')
	.then((stream)=>{
	 
		setEnabled("audio",state.audioEnabled);
		setEnabled("video",state.videoEnabled);
		if (state.publishState == "publishing")
		{
			stop();
			start();
		}

		//onMicrophoneChanged(id);
		
	});
  
}

const setCamera = (id) =>
{
  console.log("setCamera: " + id);
  if (id === 'screen')
  {
    getDisplayStream()
    .then((stream) => {
     
      setEnabled("audio",state.audioEnabled);
      setEnabled("video",state.videoEnabled);
      stream.getVideoTracks()[0].onended = function () {
        let endedState = getState();
        if (endedState.cameras.length > 0)
        {
          setCamera(endedState.cameras[0].deviceId);
        }
      }
      if (state.publishState == "publishing")
	  {
        stop();
        start();
      }
     
    });
  }
  else
  {
	let constraints = {...state.constraints};
	if (typeof constraints.video === 'boolean')
	{
	  constraints.video = {};
	}
	state.constraints.video = Object.assign({},constraints.video,{deviceId: id});
	getUserMedia('video')
	.then((stream) => {

	setEnabled("audio",state.audioEnabled);
	setEnabled("video",state.videoEnabled);
	if (state.publishState == "publishing")
	{
	  stop();
	  start();
	}

	});
   
  }
}


const log = msg => {
    document.getElementById('logs').innerHTML += msg + '<br>'
}  
const clearLogs = () => {
    $("#logs").empty();
  
}
  
